/*
 * MIT License
 *
 * Copyright (c) 2025 Quantum Unit Solutions
 * Author: David Meikle
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

import { Injectable, NestInterceptor, ExecutionContext, CallHandler } from '@nestjs/common';
import { Observable, of } from 'rxjs';
import { tap } from 'rxjs/operators';
import { Reflector } from '@nestjs/core';
import {StandaloneCachingService} from "./standalone-caching.service";

export const CACHEABLE_KEY: string = 'isCacheable';

@Injectable()
export class CacheInterceptor implements NestInterceptor {
  constructor(
      private readonly reflector: Reflector,
      private readonly cacheService: StandaloneCachingService,
  ) {}

  async intercept(context: ExecutionContext, next: CallHandler): Promise<Observable<any>> {
    const cacheMetadata: { ttlSeconds?: number } | undefined = this.reflector.get<{ ttlSeconds?: number }>(CACHEABLE_KEY, context.getHandler());

    if (cacheMetadata === undefined) {
      return next.handle();
    }

    const request: any = context.switchToHttp().getRequest();
    const cacheKey: string = this.generateCacheKey(request);

    try {
      const cachedResponse: string | null = await this.cacheService.get(cacheKey);
      if (cachedResponse) {
        return of(JSON.parse(cachedResponse)); // Return cached response if it exists
      }
    } catch (error) {
      // Handle cache retrieval error
      console.error('Cache retrieval error:', error);
    }

    // Handle the request and cache the response with a TTL
    return next.handle().pipe(
        tap((response: any) => {
          try {
            this.cacheService.set(cacheKey, JSON.stringify(response), cacheMetadata.ttlSeconds ?? 60); // Pass TTL to cache service
          } catch (error) {
            // Handle cache set error
            console.error('Cache set error:', error);
          }
        }),
    );
  }

  private generateCacheKey(request: any): string {
    return `${request.method}-${request.url}`;
  }
}