/*
 * MIT License
 *
 * Copyright (c) 2025 Quantum Unit Solutions
 * Author: David Meikle
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

import {Injectable, Logger} from '@nestjs/common';
import { RedisClientType, createClient } from 'redis';
import { RedisClientNotObtainedError } from './exceptions/redis-client-not-obtained.error';

@Injectable()
export class StandaloneCachingService {
  private _client: RedisClientType | null = null;
  private readonly logger = new Logger(this.constructor.name);
  private readonly url: string = process.env.REDIS_SERVER ?? 'redis://localhost:6379';

  constructor() {
    this._client = createClient({ url: this.url });
    this._client.connect().catch(err => this.logger.error('Redis connection error', err));
  }

  private getClient(): RedisClientType {
    if (!this._client) {
      throw new RedisClientNotObtainedError();
    }
    return this._client;
  }

  async get(key: string): Promise<any> {
    const client = this.getClient();

    if (await client.exists(key)) {
      const cacheResult = await client.get(key);
      if (cacheResult) {
        return JSON.parse(cacheResult);
      }
    }

    return null;
  }

  async set(key: string, value: any, ttl: number): Promise<void> {
    const client = this.getClient();
    await client.setEx(key, ttl, JSON.stringify(value));
  }
}